#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on 21.02.2020

Here, a set of useful postprocessing functions
(for ptychography) will be collected.

@author: Jacob Seifert  - j.seifert@uu.nl
"""

#%% Imports

import matplotlib.pyplot as plt
import numpy as np
from scipy import signal


#%% Functions

def subpixel_shift(image, dx, dy):
    ''' Performs a shift of 'image' by float number distances 'dx' and 'dy'.
            image - a 2D numpy array
            dx    - shift downwards (in pixel values)
            dy    - shift towards the right (in pixel values)
    '''
    from scipy.ndimage import fourier_shift
    image = np.fft.fft2(image)
    image = fourier_shift(image, shift=(dx, dy))
    image = np.fft.ifft2(image)
    return image.real


def get_center_of_2Dgaussian(image):
    import scipy.optimize as opt
    import numpy as np
    import matplotlib.pyplot as plt


    def twoD_Gaussian(x_y, amplitude, xo, yo, sigma_x, sigma_y, theta, offset):
        x, y = x_y
        xo = float(xo)
        yo = float(yo)
        a = (np.cos(theta)**2)/(2*sigma_x**2) + (np.sin(theta)**2)/(2*sigma_y**2)
        b = -(np.sin(2*theta))/(4*sigma_x**2) + (np.sin(2*theta))/(4*sigma_y**2)
        c = (np.sin(theta)**2)/(2*sigma_x**2) + (np.cos(theta)**2)/(2*sigma_y**2)
        g = offset + amplitude*np.exp( - (a*((x-xo)**2) + 2*b*(x-xo)*(y-yo)
                                + c*((y-yo)**2)))
        return g.ravel()

    # Preparations

    image_size = image.shape[0]  # only square images allowed for now

    # Create x and y indices
    x = np.linspace(0, image_size, image_size)
    y = np.linspace(0, image_size, image_size)
    x, y = np.meshgrid(x, y)

    #create data
    #data = twoD_Gaussian((x, y), 10, 300, 312, 300, 300, 0, 10)
    data = image.ravel()

    # # plot twoD_Gaussian data generated above
    # plt.figure()
    # plt.imshow(data.reshape(image_size, image_size))
    # plt.colorbar()

    # Fitting

    # parameter structure: Amplitude, X, Y, sigma X, sigma Y, 0, offset
    initial_guess = (1, np.int_(image_size/2), np.int_(image_size/2),
                     np.int_(image_size/4), np.int_(image_size/4), 0, 10)

    # data_noisy = data + 0.2*np.random.normal(size=data.shape)

    popt, pcov = opt.curve_fit(twoD_Gaussian, (x, y), data, p0=initial_guess)

    # Plotting fitting result

    data_fitted = twoD_Gaussian((x, y), *popt)
    data_fitted = data_fitted.reshape(image_size, image_size)
    data= data.reshape(image_size, image_size)

    # plt.imshow(data)
    # plt.contour(data_fitted, cmap=plt.cm.Reds, linewidths = 0.5)
    # plt.colorbar()
    # plt.title('Gaussian fit (contours)')
    # plt.show()
    # print('Fit parameters:\nAmplitude = {}\nX = {}, Y = {}\
    #       \nSigmaX = {}, SigmaY = {}\
    #       \nOffset = {}'.format(popt[0], popt[2], popt[1], popt[3],
    #                             popt[4], popt[6]))

    x_fit = popt[2]
    y_fit = popt[1]

    return x_fit, y_fit


def complex_correlation(a, b):
    ''' Returns the correlation magnitude of complex input 2D-arrays. '''
    field_correlation = np.dot(np.conjugate(a.flatten()), b.flatten()) / \
                               (np.linalg.norm(a)*np.linalg.norm(b))
    return np.abs(field_correlation)


def FRC(img1, img2, dl=1, drift=True, plotting=True, kernel_size=0):
    ''' Calculate the Fourier Ring Correlation of two images with independent
        noise realizations.

    inputs:
        - image 1 and image 2: must be squared and of identical size
        - dl: pixel pitch in micrometer
        - drift (boolian): corrects for drift between img1 and img2
        - plotting (boolian): creates a plot using matplotlib
        - kernel_size: integer. Determines the hamming window size
                       for apodization. If zero, no apodization is applied.
    outputs:
        - ring_frequencies: array with spatial frequencies of the rings
        - f_max: Nyquist frequency determined by pixel pitch
        - T_half_bit_smoothed: curve of the 1/2-bit threshold
        - frc: the final Fourier Ring Correlation
    source:
        https://www.osapublishing.org/optica/abstract.cfm?uri=optica-5-1-32
    '''
    # Drift correction to avoid phase ramps in the Fourier transforms
    if drift == True:
        img1, img2 = drift_correction(img1, img2)


    if kernel_size > 0:
        hamming_exponent = 1
        img1 = hamming_apodization(img1, kernel_size=kernel_size,
                                    hamming_exponent=hamming_exponent)
        img2 = hamming_apodization(img2, kernel_size=kernel_size,
                                    hamming_exponent=hamming_exponent)

    # Fourier transforming images
    FT_img1 = np.fft.fftshift(np.fft.fft2(img1))
    FT_img2 = np.fft.fftshift(np.fft.fft2(img2))

    # Frequency calculations
    f_max = 1/(2 * dl)            # maximum frequency [1/um]
    N = img1.shape[0]             # pixel count along edge
    delta_f = 1/(dl * N)          # frequency steps in Fourier domain
    f = np.linspace(-f_max, f_max-1/(N*dl), N)  # frequency vector
    fX, fY = np.meshgrid(f, f)
    f_map = np.sqrt(fX**2+fY**2)  # map that contains all frequencies that
                                  # correspond to the pixel location in the
                                  # Fourier domain

    # Two more frequency steps prevent sharp drop to 0 at Nyquist frequency:
    ring_frequencies = np.linspace(0, f_max+2*delta_f,
                                   np.int(f_max/delta_f + 3))

    # Perform FRC
    N_rings = len(ring_frequencies)
    n_q = np.zeros(N_rings)       # number of pixels contained
                                  # in Fourier ring of radius q
    frc = np.zeros(N_rings)       # Fourier Ring Correlation
    for q in range(N_rings - 1):
        # Find rings masks based on f_map and ring_frequencies
        mask_lower = f_map >= ring_frequencies[q]
        mask_upper = f_map <= ring_frequencies[q+1]
        mask = np.logical_and(mask_lower, mask_upper)
        n_q[q+1] = np.sum(mask)

        # Use the mask to select the values within the ring and correlate
        selected_frequencies1 = FT_img1[mask]
        selected_frequencies2 = FT_img2[mask]
        frc[q] = complex_correlation(selected_frequencies1,
                                     selected_frequencies2)

    # 1/2-bit resolution threshold
    # (source: https://linkinghub.elsevier.com/retrieve/pii/S1047847705001292
    n_q[n_q==0] = 1e-16   # avoid division by zero
    T_half_bit = (0.2071 + 1.9102 * 1/np.sqrt(n_q)) / \
                 (1.2071 + 0.9102 * 1/np.sqrt(n_q))

    ### Plots
    # smooth the 1/2-bit curve
    w_length = 5
    window = np.ones(w_length)
    T_half_bit_smoothed = np.convolve(window/np.sum(window),
                                      T_half_bit, mode='same')
    T_half_bit_smoothed[0:w_length] = T_half_bit[0:w_length]
    T_half_bit_smoothed[-w_length:] = T_half_bit[-w_length:]

    # Plotting
    if plotting == True:
        lw = 1.8          # linewidth

        plt.plot(ring_frequencies/f_max, np.abs(frc), '-', linewidth=lw,
                  label='Data', alpha=0.8)
        plt.plot(ring_frequencies/f_max, T_half_bit_smoothed, '--',
                  color='0.2', label='1/2-bit threshold', linewidth=lw)
        plt.grid()
        plt.legend(frameon=True, shadow=True, framealpha=0.9, facecolor='white')

        plt.xlim([0, 1])
        plt.ylim([0, 1])
        plt.ylabel('FRC')
        plt.xlabel('Nyquist frequency')
        plt.show()

    return ring_frequencies, f_max, T_half_bit_smoothed, frc


def drift_correction(img1, img2):
    ''' Performs drift correction by finding the highest cross-correlation
        between the two images img1 and img2.
    '''

    # Fourier transforming images
    FT_img1 = np.fft.fftshift(np.fft.fft2(img1))
    FT_img2 = np.fft.fftshift(np.fft.fft2(img2))

    cross_correlation = np.fft.ifftshift(np.fft.ifft2(FT_img1 *
                                                      np.conjugate(FT_img2)))
    cross_correlation = np.abs(cross_correlation)   # only magnitudes are
                                                    # relevant

    # Maximum and Center of mass
    from scipy.ndimage.measurements import center_of_mass
    from scipy.ndimage import fourier_shift

    com_ROI = 8   # ROI around the maximum
    maximum = np.where(cross_correlation == np.amax(cross_correlation))

    crop = cross_correlation[int(maximum[0]-com_ROI):
                             int(maximum[0]+com_ROI+1),
                             int(maximum[1]-com_ROI):
                             int(maximum[1]+com_ROI+1)]
    crop -= np.min(crop)
    crop /= np.max(crop)

    # fitting to 2D Gaussian and correcting half-pixel fitting offset
    fit_results = get_center_of_2Dgaussian(crop)
    x_correction = fit_results[0] - 0.5 - com_ROI
    y_correction = fit_results[1] - 0.5 - com_ROI

    # estimating total drift offsets (of img2)
    image_size = img1.shape[0]
    x_drift = -(maximum[0] + x_correction - image_size/2)
    y_drift = -(maximum[1] + y_correction - image_size/2)

    # correcting drift
    img2 = subpixel_shift(img2, -x_drift, -y_drift)

    return img1, img2

def hamming_apodization(field, kernel_size=5, hamming_exponent=1):
    # Generate 2D Hamm window and convolve it with the circle
    window = signal.hamming(kernel_size)**hamming_exponent
    window_2D = np.outer(window, window)

    res = signal.convolve2d(field, window_2D, mode='same')

    return res


#%% TESTING

testing_FRC = False

if testing_FRC:

    # Generate images
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.mathtext as mathtext
    parser = mathtext.MathTextParser("Bitmap")
    rgba1, depth1 = parser.to_rgba(
        r'$s$', color='black', fontsize=120, dpi=200)

    edge_length = 1024
    inputfield=np.zeros([edge_length, edge_length])
    textmask=rgba1[:,:,3]
    xwidth=textmask.shape[0]
    ywidth=textmask.shape[1]
    xpos=(edge_length-xwidth)//2
    ypos=(edge_length-ywidth)//2
    inputfield[xpos:xpos+xwidth,ypos:ypos+ywidth]=textmask

    # inputfield = hamming_apodization(inputfield)
    inputfield /= np.max(inputfield)

    factor = 10

    img1 = np.random.poisson(factor * inputfield) #+ np.random.normal(loc=1, scale=3, size=inputfield.shape)
    img2 = np.random.poisson(factor * inputfield)

    # Parameters
    dl = 10  # pixel pitch in mirometer

    FRC(img1, img2, dl)























# EOF