#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Library for importing ptychographical data sets into PtychoKeras.

@authors: Jacob Seifert - j.seifert@uu.nl
          Dorian Bouchet
          Allard Mosk
"""

import numpy as np

def get_data(data_string, truncation=0, propagation_mode='aspw'):
    """ Returns diffraction data and corresponding positions
        for ptychokeras.
          - data_string: can be used to reference to data by a string
          - truncation: can be used to truncate data sets for faster
                        execution. Useful for testing and prototyping.
                        0 means no truncation.
          - positions must be in mm units

        Required data array shapes:
          - diffraction images: (pixel number in x, pixel number in y,
                                 number of diffraction patterns)
          - background: same as diffraction patterns
          - positions: (number of diffraction patterns, 2 [for x and y])
    """

    # Check for known strings,

    if data_string == 'testmeasurement2020':
        """ Some measurement """
        print('Loading data...')
        data = np.load('../ptycho_data/some_ptychoMeasurement.npz')
        images = data.f.diffractionPatterns
        pos = data.f.coordinates.T     # in mm
        bg = data.f.background

    elif data_string == 'test_data':
        """ Default to dest directory and load synthetic test data"""
        print('Loading test data...')
        data = np.load('./ptycho_data/images.npz')
        images = data.f.arr_0
        data2 = np.load('./ptycho_data/coordinates.npz')
        pos = data2.f.arr_1.T

    else:
        import sys
        sys.exit("No valid data set specified. Check data string and paths.")

    # Truncating dataset
    if truncation > 0:
        images = images[:, :, 0:truncation]
        pos = pos[0:truncation, :]

    # Removing background
    if 'bg' in locals():
        for i in range(images.shape[2]):
            images[:, :, i] -= bg
            images[images < 0] = 0

    # Normalization
    images /= np.max(images)

    # Performing FFTshift (for Fraunhofer diffraction)
    if propagation_mode.lower() == "fraunhofer":
        for i in range(images.shape[2]):
            images[:, :, i] = np.fft.fftshift(images[:, :, i])

    return images, pos, data


# EOF