#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Library for ptychokeras

@authors: Jacob Seifert  - j.seifert@uu.nl
          Dorian Bouchet
          Allard Mosk
"""

#%% Imports

import numpy as np
import tensorflow as tf
from tensorflow.keras import regularizers
from tensorflow.keras import backend as K
from functions.ptycho_lib import fourier_propagator_precalculate

#%% Layer definitions

# Allocate pixel numbers which are later defined in ptycho_lib.build_ADP_model
NpixX = np.empty(1)
NpixY = np.empty(1)
m_factor = np.empty(1)

# Keras layers:

class Propagation_Layer_Fraunhofer(tf.keras.layers.Layer):
    """ Propagates the field using the Fraunhofer approximation.
        See Introduction to Fourier Optics, J. Goodman """
    # very standard initializer
    def __init__(self, **kwargs):
        super(Propagation_Layer_Fraunhofer, self).__init__(**kwargs)

    # this function is called when the layer is added. It initializes the phases to random values.
    def build(self, input_shape):
        # Create an (un)trainable weight variable for this layer.
        super(Propagation_Layer_Fraunhofer, self).build(input_shape)  # Be sure to call this at the end

    # This function actually does the work.
    def call(self, x):
        complex_x = tf.complex(x[:,0,...], x[:,1,...])
        complex_result = tf.signal.fft2d(complex_x) / NpixX
        result = tf.stack([tf.math.real(complex_result),
                           tf.math.imag(complex_result)],1)
        return result

    # The layer does not change the shape of the field tensor.
    def compute_output_shape(self, input_shape):
        return input_shape


class Propagation_Layer_Fourier(tf.keras.layers.Layer):
    """ Propagates the field using the angular spectrum method.
        See Introduction to Fourier Optics, J. Goodman """
    # very standard initializer
    def __init__(self, **kwargs):
        super(Propagation_Layer_Fourier, self).__init__(**kwargs)

    # this function is called when the layer is added. It initializes the phases to random values.
    def build(self, input_shape):
        # Create an (un)trainable weight variable for this layer.
        self.Fourierpropagator = self.add_weight(name='Fourierpropagator',
                                      shape=(2, input_shape[2], input_shape[3]),
                                      initializer='uniform',
                                      trainable=False)
        super(Propagation_Layer_Fourier, self).build(input_shape)  # Be sure to call this at the end

    # This function actually does the work.
    def call(self, x):
        complex_x=tf.complex(x[:,0,...], x[:,1,...])
        complex_fourier_propagator=tf.complex(self.Fourierpropagator[0,...], self.Fourierpropagator[1,...])
        complex_xfourier=tf.signal.fft2d(complex_x)
        complex_result=tf.signal.ifft2d(tf.multiply(complex_xfourier,complex_fourier_propagator))
        result=tf.stack([tf.math.real(complex_result),
                         tf.math.imag(complex_result)],1)
        return result

    # The layer does not change the shape of the field tensor.
    def compute_output_shape(self, input_shape):
        return input_shape


class Propagation_Layer_Fourier_with_axial_correction(tf.keras.layers.Layer):
    """ Propagates the field using the Fraunhofer approximation.
        This Layer allows for axial correction of the object-detector distance.
        See Introduction to Fourier Optics, J. Goodman """

    # very standard initializer
    def __init__(self, **kwargs):
        super(Propagation_Layer_Fourier_with_axial_correction, self).__init__(**kwargs)

    # this function is called when the layer is added. It initializes the phases to random values.
    def build(self, input_shape):
        print("input shape is", input_shape)
        # Create an (un)trainable weight variable for this layer.
        self.axialCorrection = self.add_weight(name='axialCorrection',
                                       shape=(1,),
                                       initializer='zeros',
                                       trainable=True)
        self.Fourierpropagator = self.add_weight(name='Fourierpropagator',
                                      shape=(2, input_shape[2], input_shape[3]),
                                      initializer='zeros',
                                      trainable=False)
        self.gamma = self.add_weight(name='gamma',
                                     shape=(2, input_shape[2], input_shape[3]),
                                     initializer='zeros',
                                     trainable=False)
        self.axialScaling = self.add_weight(name='axialScaling',
                                     shape=(1,),
                                     initializer='zeros',
                                     trainable=False)
        super(Propagation_Layer_Fourier_with_axial_correction, self).build(input_shape)  # Be sure to call this at the end

    # This function actually does the work.
    def call(self, x):
        complex_x=tf.complex(x[:,0,...], x[:,1,...])
        complex_fourier_propagator=tf.complex(self.Fourierpropagator[0,...], self.Fourierpropagator[1,...])
        complex_gamma = tf.complex(self.gamma[0,...], self.gamma[1,...])
        complex_axialCorrection = tf.complex(self.axialCorrection \
                                             * self.axialScaling, 0.0)

        # correcting the axial distance
        corrected_propagator = tf.multiply(complex_fourier_propagator,
                                           tf.math.exp(1j * complex_gamma * \
                                                       complex_axialCorrection))

        complex_xfourier=tf.signal.fft2d(complex_x)
        complex_result=tf.signal.ifft2d(tf.multiply(complex_xfourier,
                                                      corrected_propagator))
        result=tf.stack([tf.math.real(complex_result),
                         tf.math.imag(complex_result)],1)
        return result

    # The layer does not change the shape of the field tensor.
    def compute_output_shape(self, input_shape):
        return input_shape


class Sample_Layer(tf.keras.layers.Layer):
    """ Calculates the field behind the object using
        the projection approximation (thin object approximation).
    """
    #very standard initializer
    def __init__(self,
                 kernel_regularizer=None,
                 **kwargs):
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        super(Sample_Layer, self).__init__(**kwargs)

    # this function is called when the layer is added.
    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        # for real-space parametrization of object
        self.SampleGuess= self.add_weight(name='SampleGuess',
                                      shape=(2, NpixX, NpixY),
                                      initializer='uniform',
                                      #regularizer=self.kernel_regularizer,
                                      #constraint=self.kernel_constraint,
                                      trainable=True)

        self.InputField= self.add_weight(name='InputField',
                                      shape=(2, NpixX, NpixY),
                                      initializer='zeros',
                                      #regularizer=self.kernel_regularizer,
                                      trainable=True)  # trainability of the probe
        super(Sample_Layer, self).build(input_shape)  # Be sure to call this at the end

    #this function actually does the work.
    def call(self, pos):
        # Recovering the complex variables
        complex_input=tf.complex(self.InputField[0,:,:], self.InputField[1,:,:])
        complex_guess=tf.complex(self.SampleGuess[0,:,:], self.SampleGuess[1,:,:])

        complex_shiftedguess=tf.map_fn(lambda posn: tf.roll(complex_guess,
                                                            -posn, axis=[0,1]),
                                       pos, dtype=tf.complex64)
        complex_result = tf.multiply(complex_shiftedguess, complex_input)
        result = tf.stack([tf.math.real(complex_result),
                           tf.math.imag(complex_result)], 1)
        return result

    # The layer always has output shjape (None,2,Nx,Ny)
    def compute_output_shape(self, input_shape):
        return (None,2,NpixX,NpixY)


class Sample_Layer_exponentialForm(tf.keras.layers.Layer):
    """ Calculates the field behind the object using
        the projection approximation (thin object approximation).
        Uses the exponential form to describe complex numbers.
    """
    #very standard initializer
    def __init__(self,
                 kernel_regularizer=None,
                 **kwargs):
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        super(Sample_Layer_exponentialForm, self).__init__(**kwargs)

    # this function is called when the layer is added.
    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        # for real-space parametrization of object
        self.SampleGuess= self.add_weight(name='SampleGuess',
                                      shape=(2, NpixX, NpixY),
                                      initializer='zeros',
                                      #regularizer=self.kernel_regularizer,
                                      #constraint=self.kernel_constraint,
                                      trainable=True)

        self.InputField= self.add_weight(name='InputField',
                                      shape=(2, NpixX, NpixY),
                                      initializer='zeros',
                                      #regularizer=self.kernel_regularizer,
                                      trainable=True)  # trainability of the probe
        super(Sample_Layer_exponentialForm, self).build(input_shape)  # Be sure to call this at the end

    #this function actually does the work.
    def call(self, pos):
        # Recovering the complex variables
        
        input_magnitude = tf.complex(self.InputField[0,:,:], 
                                     tf.zeros_like(self.InputField[0,:,:]))
        input_phase = tf.complex(self.InputField[1,:,:], 
                                     tf.zeros_like(self.InputField[1,:,:]))
        complex_input=tf.multiply(input_magnitude, 
                                  tf.math.exp(1j * input_phase))
        
        object_magnitude = tf.complex(self.SampleGuess[0,:,:], 
                                     tf.zeros_like(self.SampleGuess[0,:,:]))
        object_phase = tf.complex(self.SampleGuess[1,:,:], 
                                     tf.zeros_like(self.SampleGuess[1,:,:]))
        complex_guess = tf.multiply(object_magnitude, 
                                    tf.math.exp(1j * object_phase))
        
        # Shifting the object
        complex_shiftedguess=tf.map_fn(lambda posn: tf.roll(complex_guess,
                                                            -posn, axis=[0,1]),
                                       pos, dtype=tf.complex64)
        complex_result = tf.multiply(complex_shiftedguess, complex_input)
        result = tf.stack([tf.math.real(complex_result),
                           tf.math.imag(complex_result)], 1)
        return result

    # The layer always has output shjape (None,2,Nx,Ny)
    def compute_output_shape(self, input_shape):
        return (None,2,NpixX,NpixY)



class Sample_Layer_sliced(tf.keras.layers.Layer):
    """ Calculates the field behind the object using
        the projection approximation (thin object approximation).

        This function updates only slices of the total object FOV
        and requires a multiplicity factor for the ration of the total FOV to
        the slice: m_factor

        This Layer is useful for reconstructions with
        very large space-bandwith products.
    """

    #very standard initializer
    def __init__(self,
                 kernel_regularizer=None,
                 **kwargs):
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        super(Sample_Layer_sliced, self).__init__(**kwargs)

    # this function is called when the layer is added.
    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        # for real-space parametrization of object
        self.SampleGuess= self.add_weight(name='SampleGuess',
                                      shape=(2, m_factor * NpixX, m_factor * NpixY),
                                      initializer='ones',
                                      #regularizer=self.kernel_regularizer,
                                      #constraint=self.kernel_constraint,
                                      trainable=True)

        self.InputField= self.add_weight(name='InputField',
                                      shape=(2, NpixX, NpixY),
                                      initializer='zeros',
                                      #regularizer=self.kernel_regularizer,
                                      trainable=True)  # trainability of the probe
        super(Sample_Layer_sliced, self).build(input_shape)  # Be sure to call this at the end

    #this function actually does the work.
    def call(self, pos):
        # Recovering the complex variables
        complex_input=tf.complex(self.InputField[0,:,:], self.InputField[1,:,:])
        complex_guess=tf.complex(self.SampleGuess[0,:,:], self.SampleGuess[1,:,:])

        complex_shiftedguess=tf.map_fn(lambda posn: tf.roll(complex_guess, -posn, axis=[0,1]), pos, dtype=tf.complex64)
        complex_sliced_guess = complex_shiftedguess[:, m_factor//2*NpixX:(m_factor//2+1)*NpixX, m_factor//2*NpixX:(m_factor//2+1)*NpixX]

        print(complex_guess.shape, complex_shiftedguess.shape, complex_sliced_guess.shape)

#        # This should work and speed up things, but it doesn't for now
#        # e.g, for m_factor = 7:
#        row = -pos[0][1]
#        col = -pos[0][0]
#        complex_sliced_guess = complex_guess[3*NpixX+row:4*NpixX+row,
#                                             3*NpixY+col:4*NpixY+col]

        complex_result = tf.multiply(complex_sliced_guess, complex_input)
        result = tf.stack([tf.math.real(complex_result),
                           tf.math.imag(complex_result)], 1)
        return result

    # The layer always has output shjape (None,2,Nx,Ny)
    def compute_output_shape(self, input_shape):
        return (None, 2, NpixX, NpixY)


def To_intensity(x):
    """ This lambda layer function converts a field to an intensity. """
    return K.sum(tf.square(x), 1, keepdims=True)


#%% Helper functions (e.g., for setting layer weights)

def Propagation_Layer_Setweights(layer, k0, z, dl, NA_max):
    """ Helper function for setting the weights of the propagation layer.

        Input:
            - layer: Keras layer
            - k0: angular wavenumber
            - z: camera-detector distance
            - dl: pixel pitch
            - NA_max: numerical aperture
    """
    weights_temp = layer.get_weights()
    theshape = weights_temp[0].shape
    precalc = fourier_propagator_precalculate(theshape[1:], k0, z,
                                              dl, NA_max=NA_max)

    weights_temp[0][0,:,:] = precalc.real.astype('float32')
    weights_temp[0][1,:,:] = precalc.imag.astype('float32')

    layer.set_weights(weights_temp)


def Propagation_Layer_Setweights_with_axial_correction(layer, k0, z, dl,
                                                       NA_max, gamma,
                                                       axialScaling):
    """ Helper function for setting the weights of the propagation layer.
        Includes gamma for axial correction.

        Input:
            - layer: Keras layer
            - k0: angular wavenumber
            - z: camera-detector distance
            - dl: pixel pitch
            - NA_max: numerical aperture
            - gamma: gamma variable for wave propagation (see. J. Goodman)
            - axialScaling: scales the gradient step size for axial correction
    """
    weights_temp = layer.get_weights()
    theshape = weights_temp[1].shape
    precalc = fourier_propagator_precalculate(theshape[1:], k0, z,
                                              dl, NA_max=NA_max)

    weights_temp[1][0,:,:] = precalc.real.astype('float32')
    weights_temp[1][1,:,:] = precalc.imag.astype('float32')

    # gamma
    weights_temp[2][0, :, :] = gamma.real.astype('float32')
    weights_temp[2][1, :, :] = gamma.imag.astype('float32')

    # axial scaling factor
    weights_temp[3][0] = axialScaling

    layer.set_weights(weights_temp)


def To_intensity_output_shape(input_shape):
    """ Helper function that calculates the output shape. """
    return (input_shape[0],2,input_shape[1],input_shape[2])


def Sample_Layer_Setweights(layer, inputfield):
    """ Helper function for setting the weights of the initial object.

        Input:
            - layer: Keras layer
            - inputfield: inital probe guess
    """
    weights_temp=layer.get_weights()
    weights_temp[0] = 1 * np.ones_like(weights_temp[0])  # this is the object

    # generate random uniformly distributed comlex numbers in a unit circle
    if False:
        objectshape = weights_temp[0].shape
        total_pixels = objectshape[1] * objectshape[2]
        magnitudes = np.random.random_sample(total_pixels)
        angles = 2 * np.pi * np.random.random_sample(total_pixels)
        k = 0
        complex_uniform = magnitudes * np.exp(-1j*angles)
        for i in range(objectshape[1]):
            for j in range(objectshape[2]):
                    weights_temp[0][0,i,j] = complex_uniform[k].real.astype('float32')
                    weights_temp[0][1,i,j] = complex_uniform[k].imag.astype('float32')                    
                    k += 1

    # Set the initial probe function
    theshape=weights_temp[1].shape;
    for i in range(theshape[1]):
        for j in range(theshape[2]):
                weights_temp[1][0,i,j] = inputfield[i,j].real.astype('float32')
                weights_temp[1][1,i,j] = inputfield[i,j].imag.astype('float32')
                # weights_temp[1][0,i,j] = np.abs(inputfield[i,j])
                # weights_temp[1][1,i,j] = np.angle(inputfield[i,j])         
    layer.set_weights(weights_temp)