#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Library for ptychokeras

@authors: Jacob Seifert  - j.seifert@uu.nl
          Dorian Bouchet
          Allard Mosk
"""

import numpy as np


def build_ADP_model(inputfield, parameters,
                    propagation_mode="aspw", axial_correction=False):
    """ Serves as a function to modularize our layers for ADP.
        Input:
            - inputfield: first guess for the probe function
            - parameter dictionary (from ptychokeras.py)
            - propagation mode string (aspw or fraunhofer)
        Output:
            - Keras model of the physics forward model in ptychography

        To build a model that allows fitting of the axial distance, replace
        Sample_Layer_Setweights() with Sample_Layer_Setweights_with_axial_correction()
        and
        Propagation_Layer_Setweights() with Propagation_Layer_Setweights_with_axial_correction()
    """

    # Loading Libraries (and setting pixel numbers in layers library)
    import tensorflow as tf
    import functions.ptycho_layers as ptycho_layers
    ptycho_layers.NpixX = parameters.get('NpixX')
    ptycho_layers.NpixY = parameters.get('NpixX')
    
    ### Building the model
    model = tf.keras.Sequential()

    # Scanning positions
    model.add(tf.keras.layers.InputLayer(input_shape=(2,), dtype="int32"))

    # Object layer
    if propagation_mode.lower() == "aspw":
        # model.add(ptycho_layers.Sample_Layer_exponentialForm())
        model.add(ptycho_layers.Sample_Layer())

    elif propagation_mode.lower() == "fraunhofer":
        # defining total object dimenstion with m_factor 
        ptycho_layers.m_factor = 3  # total object size = m_factor * slice size
        model.add(ptycho_layers.Sample_Layer_sliced())

    # Writing the initial probe guess to the object layer
    ptycho_layers.Sample_Layer_Setweights(model.layers[-1], inputfield)

    # Propagation layer
    if propagation_mode.lower() == "aspw":
        if axial_correction:
            gamma = get_gamma(parameters.get('shape'),
                              parameters.get('k0'),
                              parameters.get('dl'))
            model.add(ptycho_layers.Propagation_Layer_Fourier_with_axial_correction())
            ptycho_layers.Propagation_Layer_Setweights_with_axial_correction(model.layers[-1],
                                                                            parameters.get('k0'),
                                                                            parameters.get('zscreen'),
                                                                            parameters.get('dl'),
                                                                            parameters.get('NA_max'),
                                                                            gamma,
                                                                            parameters.get('axial_scaling'))
        else:
            model.add(ptycho_layers.Propagation_Layer_Fourier())
            ptycho_layers.Propagation_Layer_Setweights(model.layers[-1],
                                                    parameters.get('k0'),
                                                    parameters.get('zscreen'),
                                                    parameters.get('dl'),
                                                    parameters.get('NA_max'))

    elif propagation_mode.lower() == "fraunhofer":
        model.add(ptycho_layers.Propagation_Layer_Fraunhofer())

    # Intensity layer in the detection plane
    model.add(tf.keras.layers.Lambda(ptycho_layers.To_intensity))
    model.layers[-1].trainable = False

    return model


def fourier_propagator_precalculate(size, k0, z, dl, NA_max=1):
    """ Precalculates the propagator for free space propagation from
        the object to the detection plane using the angular spectrum method.

        Input:
            - size: pixel dimension of the propagated field (x, y)
            - k0: angular wavenumber
            - z: object-detection plane distance
            - dl: sampling pitch
            - NA_max: numerical aperture
        Output:
            - angular spectrum method propagator
    """

    Lx = size[0]*dl
    Ly = size[1]*dl

    freqX = np.fft.ifftshift(np.linspace(-1/(2*dl),1/(2*dl)-1/Lx, num=size[0]))
    freqY = np.fft.ifftshift(np.linspace(-1/(2*dl),1/(2*dl)-1/Ly, num=size[1]))
    alpha, beta = 2*np.pi * np.asarray(np.meshgrid(freqY, freqX))
    gamma = np.power((k0**2 - alpha**2 - beta**2 + 0j), 0.5)

    # Implementing a slight smoothing filter of the propagator
    filter_init = np.zeros(size, dtype=np.complex_)
    smoothmax = 1.1
    smoothmin = 0.9
    dist2 = alpha**2 + beta**2
    x, y = np.where(smoothmin**2*NA_max**2*k0**2>alpha**2+beta**2)
    filter_init[x, y] = 1
    xgauss, ygauss = np.where(np.logical_and(smoothmin**2*NA_max**2*k0**2<dist2,smoothmax**2*NA_max**2*k0**2>dist2))
    filter_init[xgauss, ygauss] = np.exp(-(dist2[xgauss,ygauss]-NA_max**2*k0**2*smoothmin**2)/(k0**2*NA_max**2*(smoothmax**2-smoothmin**2)*0.5))

    # Building the propagator
    propagator = np.exp(1j*gamma*z) * filter_init

    return propagator


def get_gamma(size, k0, dl):
    """ Returns gamma used for axial distance correction """
    Lx=size[0]*dl
    Ly=size[1]*dl

    freqX=np.fft.ifftshift(np.linspace(-1/(2*dl),1/(2*dl)-1/Lx,num=size[0]))
    freqY=np.fft.ifftshift(np.linspace(-1/(2*dl),1/(2*dl)-1/Ly,num=size[1]))
    alpha,beta=2*np.pi*np.asarray(np.meshgrid(freqY,freqX))
    gamma=np.power((k0**2-alpha**2-beta**2+0j),0.5)
    return gamma

def fourier_propagate(Einit, precalculated_propagator):
    """ Numpy free space propagation (angular spectrum method) of the field
        'Einit' using the propagator 'precalculated_propagator'.
    """
    FT_Einit = np.fft.fft2(Einit)
    FT_Efinal = precalculated_propagator * FT_Einit
    Efinal = np.fft.ifft2(FT_Efinal)

    return Efinal

def smooth_circular_aperture(radius, dl, Lx, Ly, offset, dr):
    """ Helper function to generate a pinhole to estimate the probe in a
        single aperture ptychography setup.

        Input:
            - radius: radius of the pinhole
            - dl: sampling pitch
            - Lx, Ly: extent of the view in x and y
            - offset: offset of the aperture from the image's center (x, y)
            - dr: distance over which smoothing is performed (um)
        Output:
            - smoothed pinhole
    """

    NoffX = offset[0]/dl
    NoffY = offset[1]/dl
    Nx = int(Lx/dl)
    Ny = int(Ly/dl)
    Einit = np.zeros((Ny,Nx),dtype=np.complex_)
    NpixvectX = np.linspace(0,Nx-1,num=Nx)
    NpixvectY = np.linspace(0,Ny-1,num=Ny)
    xg, yg = np.meshgrid(NpixvectX,NpixvectY)
    x, y = np.where(dl**2*((xg-(Nx/2+0.5)-NoffX)**2+(yg-(Ny/2+0.5)-NoffY)**2)<=(radius-dr)**2)
    Einit[x, y] = 1
    xgauss, ygauss = np.where(np.logical_and(dl**2*((xg-(Nx/2+0.5)-NoffX)**2+(yg-(Ny/2+0.5)-NoffY)**2)>(radius-dr)**2,dl**2*((xg-(Nx/2+0.5)-NoffX)**2+(yg-(Ny/2+0.5)-NoffY)**2)<(radius+dr)**2))
    dist = np.sqrt(dl**2*((ygauss-(Nx/2+0.5)-NoffX)**2+(xgauss-(Ny/2+0.5)-NoffY)**2))-(radius-dr)
    Einit[xgauss, ygauss] = np.exp(-dist**2/(2*(dr/2)**2))

    return Einit.transpose()


#%% Functions for visualizing complex-valued 2D images

def complex_colorize(z, rmin=False, rmax=False, hue_start=180):
    """ Returns a real RGB image base on some complex input. """
    from matplotlib.colors import hsv_to_rgb

    # get amplitude of z and limit to [rmin, rmax]
    amp = np.abs(z)
    if not rmin and not rmax:
        rmin = 0
        rmax = np.max(amp)
    amp = np.where(amp < rmin, rmin, amp)
    amp = np.where(amp > rmax, rmax, amp)
    ph = np.angle(z, deg=1) + hue_start

    # HSV are values in range [0,1]
    h = (ph % 360) / 360
    s = 0.80 * np.ones_like(h)
    v = (amp - rmin) / (rmax - rmin)
    return hsv_to_rgb(np.dstack((h, s, v)))


def viz(image, title='none', x_dim=False, y_dim=False):
    ''' Visualizes a 2D image, array etc. (can be complex) '''
    import matplotlib.pyplot as plt

    is_complex = False

    if image.dtype == 'complex128' or image.dtype == 'complex64':
        image = complex_colorize(image)
        is_complex = True

    fig, ax = plt.subplots(figsize=(8, 4), dpi=100)
    if x_dim and y_dim:
        plt.imshow(image, extent=[-y_dim/2e3, y_dim/2e3,
                                  -x_dim/2e3, x_dim/2e3])
        plt.xlabel('y (mm)')
        plt.ylabel('x (mm)')
    else:
        plt.imshow(image)
    if title != 'none':
        plt.title(title)
    if not is_complex:
        plt.colorbar()
    plt.show()

    return

# EOF