# -*- coding: utf-8 -*-
"""
Script that performs ptychography reconstruction
using automatic differentiation.
Read README.md for information about dependencies and how to run the script.

@authors: Jacob Seifert - j.seifert@uu.nl
          Dorian Bouchet
          Allard Mosk
"""

#%% Importing functions

import functions.ptycho_lib as ptycho_lib
import functions.ptycho_datasets as ptycho_data
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf

#%% Settings

# Data
data_string = 'test_data'  # if you change this string, edit ptych_datasets
data_truncation = -1       # truncate dataset for testing (-1 = no truncation)

# Reconstruction settings
epochs = 50                 # epochs
lr = 0.01                   # learning rate (Adam)
axial_correction = False    # Boolean var for fitting camera-object distance,
                            # requires propagation_mode='aspw'
                            # requires tensorflow version >= 2.2.0
propagation_mode = 'aspw'   # aspw or fraunhofer

#%% Loading data

images, pos, data_compressed = ptycho_data.get_data(data_string,
                                                    data_truncation,
                                                    propagation_mode)

#%% Experimental parameters

lambda0 = 0.561          # wavelength (um)
k0 = 2*np.pi/lambda0     # wavenumber (um^-1)
radius = 50.             # radius of pinhole (um)
dr = 30.                 # distance over which smoothing is performed (um)
dl = 6.45                # sampling in the camera plane (um)
dl_object = dl           # sampling in the object plane (um), only for aspw.
NpixX = images.shape[0]  # pixel count in x
NpixY = images.shape[1]  # pixel count in y
Lx = dl * NpixX          # field of view (um), pixel size * resolution
Ly = dl * NpixY
Npix = NpixX * NpixY     # total count of pixels

zpinhole = 93.76e3       # distance between pinhole and object (um)
zscreen = 146.00e3       # distance between object and screen (um)
probe_offset = (0, 0)
NA_max = np.min([Lx, Ly]) / (2 * zscreen)  # numerical aperture
axial_scaling_factor = np.copy(zscreen)  # rescaling of the axial correction during optimization (should be in the same order as zscreen)

# Building a parameter dictionary

parameters = {}
parameters['k0'] = k0
parameters['zscreen'] = zscreen
parameters['dl'] = dl
parameters['NA_max'] = NA_max
parameters['shape'] = (NpixX, NpixY)
parameters['axial_scaling'] = axial_scaling_factor
parameters['NpixX'] = NpixX
parameters['NpixY'] = NpixY

#%% Computing or loading the initial probe function

pinhole = ptycho_lib.smooth_circular_aperture(radius, dl, Lx, Ly,
                                              probe_offset, 15)
precalc = ptycho_lib.fourier_propagator_precalculate(pinhole.shape, k0,
                                                     zpinhole, dl,
                                                     NA_max=0.01)
inputfield = ptycho_lib.fourier_propagate(pinhole, precalc)
ptycho_lib.viz(inputfield, title='Initial probe guess')   # plotting the probe

#%% Shaping the tensors

print('Tensor shaping...')
x_train = pos[:,:] * 1e3 / dl_object   # scanning positions in pixel values
# Shifting image to center
x_train -= [np.average(x_train[:, 0]), np.average(x_train[:, 1])]
y_train = np.transpose(images, (2, 0, 1))[:, np.newaxis, :, :]
ffilter = np.fft.fftshift(ptycho_lib.fourier_propagator_precalculate((y_train[0,0].shape),
                                                          k0, 0.0, dl,
                                                          NA_max=NA_max))
y_train_f = np.fft.fftshift(np.fft.fft2(y_train),(-1,-2))
for i in range(len(y_train)):
                 img=y_train_f[i]*ffilter
                 y_train_f[i]=img
                 y_train[i]=np.abs(np.fft.ifft2(np.fft.ifftshift(img,(-1,-2))))

#%% Defining the model using Keras

print('Building Keras model...')
model = ptycho_lib.build_ADP_model(inputfield, parameters,
                                   propagation_mode, axial_correction)
model.summary()

#%% Compiling and fitting

opt = tf.keras.optimizers.Adam(lr=lr)
loss = tf.keras.losses.MSE
model.compile(optimizer=opt, loss=loss)
print('Starting ADAM optimization...')
history = model.fit(x_train, y_train,  epochs=epochs, batch_size=1, verbose=1)

#%% Plot the convergence
plt.figure(figsize=(8,4))
loss_array = np.array(history.history['loss'], dtype='float32')
plt.semilogy(loss_array)
plt.xlabel('Epoch')
plt.ylabel('Loss (MSE)')
plt.grid()
plt.show()

#%% Plot the retrieved object
complex_object=(model.layers[0].get_weights()[0][1]*1j+model.layers[0].get_weights()[0][0])
plt.figure(figsize=(15,10))
plt.subplot(121)
plt.imshow(np.abs(complex_object), extent=(0, dl_object*NpixY/1e3, 0,
           dl_object*NpixX/1e3))
plt.xlabel('[mm]')
plt.ylabel('[mm]')
plt.title('Optical density contrast')
plt.colorbar(fraction=0.046, pad=0.04)

plt.subplot(122)
plt.imshow(np.angle(complex_object), extent=(0, Ly/1e3, 0, Lx/1e3))
plt.colorbar(fraction=0.046, pad=0.04)
plt.xlabel('[mm]')
plt.ylabel('[mm]')
plt.title('Phase contrast')
#plt.clim([0, np.pi/2])
plt.show()

#%% Plot the retrieved probe
complex_probe = (model.layers[0].get_weights()[1][1]*1j+model.layers[0].get_weights()[1][0])
ptycho_lib.viz(complex_probe, title='Probe after reconstruction')


#%% Exporting results

# # create folder
# import os
# outPath = 'output'
# if not os.path.exists(outPath):
#     os.makedirs(outPath)

# # saving file
# np.savez_compressed('{}/ADP_lr{}_{}epochs.npz'.format(outPath, str(lr),
#                     str(epochs)), obj=complex_object,
#                     probe=complex_probe, loss=loss_array)

# EOF
